pub trait WidgetBehavior {
    fn click(
        &mut self,
        _widget: &mut Widget,
        _mouse_x: f32,
        _mouse_y: f32,
    ) {
        println!("Ouch ! Somebody clicked on me !");
    }

    fn draw(
        &self,
        widget: &Widget,
    );
}

pub struct Widget {
    x: f32,
    y: f32,
    width: usize,
    height: usize,
    behave: Option<Box<dyn WidgetBehavior>>,
}

impl Widget {
    pub fn new(
        x: f32,
        y: f32,
        width: usize,
        height: usize,
        behave: Box<dyn WidgetBehavior>,
    ) -> Self {
        Self {
            x,
            y,
            width,
            height,
            behave: Some(behave),
        }
    }

    pub fn click(
        &mut self,
        mouse_x: f32,
        mouse_y: f32,
    ) {
        if mouse_x >= self.x
            && mouse_x < self.x + self.width as f32
            && mouse_y >= self.y
            && mouse_y < self.y + self.height as f32
        {
            if let Some(mut behave) = self.behave.take() {
                behave.click(self, mouse_x, mouse_y);
                self.behave = Some(behave);
            }
        }
    }

    pub fn draw(&self) {
        if let Some(behave) = &self.behave {
            behave.draw(self);
        }
    }
}

pub struct Label {
    pub caption: String,
}

impl Label {
    pub fn new(caption: String) -> Self {
        Self { caption }
    }
}

impl WidgetBehavior for Label {
    fn draw(
        &self,
        widget: &Widget,
    ) {
        println!(
            "Label in {:?} of [{}x{}] with caption {:?}",
            (widget.x, widget.y),
            widget.width,
            widget.height,
            self.caption
        );
    }
}

#[derive(Default)]
pub struct Text {
    pub txt: String,
}

impl Text {
    pub fn new() -> Self {
        Self { txt: "".to_owned() }
    }
}

impl WidgetBehavior for Text {
    fn click(
        &mut self,
        _widget: &mut Widget,
        _mouse_x: f32,
        _mouse_y: f32,
    ) {
        println!("Text {:?} clicked !", self.txt);
    }

    fn draw(
        &self,
        widget: &Widget,
    ) {
        println!(
            "Text in {:?} of [{}x{}] with txt {:?}",
            (widget.x, widget.y),
            widget.width,
            widget.height,
            self.txt
        );
    }
}

pub struct Button {
    pub name: String,
}

impl Button {
    pub fn new(name: String) -> Self {
        Self { name }
    }
}

impl WidgetBehavior for Button {
    fn click(
        &mut self,
        widget: &mut Widget,
        _mouse_x: f32,
        _mouse_y: f32,
    ) {
        println!("Button {:?} clicked !", self.name);
        widget.width += 5;
        widget.height += 5;
    }

    fn draw(
        &self,
        widget: &Widget,
    ) {
        println!(
            "Button in {:?} of [{}x{}] with name {:?}",
            (widget.x, widget.y),
            widget.width,
            widget.height,
            self.name
        );
    }
}

pub struct Container {
    title: String,
    widgets: Vec<Widget>,
}

impl Container {
    pub fn new(title: String) -> Self {
        Self {
            title,
            widgets: Vec::new(),
        }
    }

    pub fn add(
        &mut self,
        widget: Widget,
    ) {
        self.widgets.push(widget);
    }
}

impl WidgetBehavior for Container {
    fn click(
        &mut self,
        widget: &mut Widget,
        mouse_x: f32,
        mouse_y: f32,
    ) {
        println!("Container with title {:?} clicked !", self.title);
        for w in self.widgets.iter_mut() {
            w.click(mouse_x - widget.x, mouse_y - widget.y);
        }
    }

    fn draw(
        &self,
        widget: &Widget,
    ) {
        println!(
            "Container in {:?} of [{}x{}] with title {:?}",
            (widget.x, widget.y),
            widget.width,
            widget.height,
            self.title
        );
        for w in self.widgets.iter() {
            w.draw();
        }
    }
}
